#from ast import pattern
import os
import sys
import getopt
import random
from threading import local
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.widgets import Slider

from numpy.random.mtrand import rand
from jmetal.algorithm.multiobjective.moead import LearningMOEAD
from jmetal.core.quality_indicator import HyperVolume
from jmetal.util.aggregative_function import WeightedSum
from jmetal.util.termination_criterion import StoppingByTime
from jmetal.operator.crossover import PMXCrossover
from jmetal.operator.extraction import MOPatternExtraction
from jmetal.operator.injection import MOPatternInjection
from jmetal.operator.mutation import PermutationSwapMutation
from jmetal.operator.localSearch import ManyOperators
from jmetal.problem.multiobjective.movrptw import MOVRPTW
from jmetal.lab.experiment import Experiment, Job, generate_summary_from_experiment
from jmetal.core.solution import Solution
 
from jmetal.lab.visualization.plotting import Plot
from jmetal.util.solution import get_non_dominated_solutions

# Current path to access instance files: os.path.join('resources', 'VRPTW_instances', 'Solomon', 'solomon_'+instanceSize, instanceType + instanceNumber)+".txt"

def read_arguments(argv):

    try:                                
        opts, args = getopt.getopt(argv, "lidhk", ["learning", "intensification", "diversification", "hypervolume", "knowledge"])
    except getopt.GetoptError:                           
        sys.exit(2)  
    
    allParameters = {}
    nbArgs = len(argv[1:])
    allParameters['injectionStrategy'] = ""
    allParameters['hvCriterion'] = False
    allParameters['learning'] = False
    allParameters['displayKnowledge'] = False
    allParameters['mutation'] = "LocalSearch"
    if opts:
        for opt in opts:
            if opt[0] in ["-i", "-d"]:
                if opt[0] == "-i":
                    allParameters['injectionStrategy'] = "intensification"
                else:
                    allParameters['injectionStrategy'] = "diversification"
            
            elif opt[0] == "-h":
                allParameters['hvCriterion'] = True
            
            elif opt[0] == "-l":
                allParameters['learning'] = True

            elif opt[0] == "-k":
                allParameters['displayKnowledge'] = True

    # Instance Parameters
    allParameters['nameAlgo'] = args[0]       
    allParameters['seed'] = int(args[1])
    allParameters['benchmark'] = args[2]
    allParameters['numberOfRun'] = int(args[3])
    allParameters['instanceNumber'] = args[4]
    allParameters['instanceType'] = args[5]
    allParameters['instanceSize'] = args[6]
    
    # Algo Parameters
    allParameters['maxTime'] = int(args[7])
    allParameters['populationSize'] = int(args[8])
    allParameters['sizeNeighborhood'] = int(args[9])
    allParameters['granularity'] = int(args[10])
    allParameters['probabilityCrossover'] = float(args[11])
    allParameters['probabilityMutation'] = float(args[12])
    
    # Learning
    allParameters['maxPatternSize'] = int(args[13])
    allParameters['patternsConsidered'] = int(args[14])
    allParameters['patternsInjected'] = int(args[15])
    allParameters['frequencyExtraction'] = float(args[16])
    allParameters['frequencyInjection'] = float(args[17])

    return allParameters
    
def configure_experiment(allParameters: dict):
    instanceNumber = allParameters['instanceNumber']
    instanceType = allParameters['instanceType']
    instanceSize = allParameters['instanceSize'] 

    if allParameters['benchmark'] == "Solomon":
        subpath = os.path.join('Solomon', 'solomon_'+instanceSize, instanceType + instanceNumber)
        instance = os.path.join('resources', 'VRPTW_instances', subpath + ".txt") 
        
    elif allParameters['benchmark'] == "Generated":
        if int(instanceNumber) < 10:
            instanceNumber = "0" + instanceNumber
        subpath = os.path.join('generated_instances', 'gen_'+instanceSize, instanceType + instanceNumber)
        instance = os.path.join('resources', 'VRPTW_instances', subpath)

    extrema = os.path.join('extrema', subpath, 'extrema.txt')
    idealPoint = []
    nadirPoint = []
    if os.path.isfile(extrema):
        with open(extrema, 'r') as file:
            lines = file.readlines()
            for line in lines:
                values = [float(i) for i in line.split(' ')[:-1]]
                idealPoint.append(values[0])
                nadirPoint.append(values[1])

    hvGoal = 2.0 # unreachable value
    if allParameters['hvCriterion']:
        # define a hypervolume to reach
        algoRef = 'MOEAD-LS'
        fileHV = os.path.join('data', algoRef, subpath, 'ALL-HYPERVOLUMES_10.tsv')
        if not os.path.isfile(fileHV):
            raise FileExistsError('The file does not exist. Specify a path to a hypervolume data.')
        allHV = []
        with open(fileHV, 'r') as file:
            lines = file.readlines()
            for line in lines:
                hv = float(line.split(' ')[0])
                allHV.append(hv)
        allHV.sort()
        hvGoal = round(0.95 * np.mean(allHV), 3)

    run = allParameters['numberOfRun']
    initial_seed = allParameters['seed']
    max_neighbours = allParameters['granularity']

    problem = MOVRPTW(max_neighbours, instance, nadirPoint, idealPoint)
    problem_tag = os.path.join(subpath, "Run"+str(run+1), "Final")

    neighbourhood_selection_probability = 1.0
    max_dico_size = 2000
    mutation = ManyOperators(probability = allParameters['probabilityMutation'])

    job = Job(
            algorithm = LearningMOEAD(
                problem = problem,
                population_size = allParameters['populationSize'],
                max_pattern_size = allParameters['maxPatternSize'],
                max_dico_size = max_dico_size,
                number_of_patterns_injected = allParameters['patternsInjected'],
                number_of_patterns_considered = allParameters['patternsConsidered'],
                mutation = mutation,
                crossover = PMXCrossover(probability = allParameters['probabilityCrossover']),
                extraction = MOPatternExtraction(probability = allParameters['frequencyExtraction']),
                injection = MOPatternInjection(probability = allParameters['frequencyInjection'], strategy = allParameters['injectionStrategy']),
                aggregative_function = WeightedSum(),
                neighbourhood_selection_probability = neighbourhood_selection_probability,
                max_number_of_replaced_solutions = 2,
                neighbor_size = allParameters['sizeNeighborhood'],
                weight_files_path = 'resources/MOEAD_weights',
                output_path = os.path.join('data', allParameters['nameAlgo'], subpath, "Run"+str(run+1), "Snapshots"),
                termination_criterion = StoppingByTime(allParameters['maxTime']),
                hypervolumeCriterion = hvGoal,
                learningOnLO = allParameters['learning']),
            algorithm_tag= allParameters['nameAlgo'],
            problem_tag=problem_tag,
            run=run,
            seed = initial_seed * (run)* 10
        )

    return job

#############################################################################################
# The following functions are used only when the knowledge is displayed after the execution #
#############################################################################################

def plot_instance(ax, problem: MOVRPTW):
    positions = problem.positions
    x = []
    y = []
    for i in range(len(positions)):
        if i%2 == 0:
            x.append(positions[i])
        else:
            y.append(positions[i])
    ax.scatter(x[0], y[0], color = 'red')
    ax.scatter(x[1:], y[1:], color = 'blue')
    return

def plot_patterns(ax, patterns, problem: MOVRPTW):
    # arcs contain the list of patterns to plot
    # a pattern is modeled as a list 
    positions = problem.positions
    frequency_max = patterns[0][1]
    for (pattern, frequency) in patterns:
        lgth = len(pattern)
        for i in range(lgth-1):
            a = pattern[i]
            b = pattern[i+1]
            xa = positions[2*a]
            ya = positions[2*a+1]
            xb = positions[2*b]
            yb = positions[2*b+1]
            ax.plot([xa, xb], [ya, yb], color = 'black', linewidth = frequency/frequency_max) # could add a linewidth depending on the frequency of the pattern
    return

def computePatterns(group, sizePattern, numberPatterns, knowledge):
    currentKnowledge = knowledge[group][sizePattern]
    listPatterns = []
    sorted_knowledge = sorted(currentKnowledge.items(), key=lambda t: -t[1])
    relevant_patterns = sorted_knowledge[:numberPatterns]
    for element in relevant_patterns:
        pattern = [int(i) for i in element[0].split(' ')[:-1]]
        listPatterns.append((pattern, element[1]))
    return listPatterns

def updateFigure(val):
    sizePattern = size_slider.val
    group = group_slider.val
    nbPatterns = nbPattern_slider.val
    listPatterns = computePatterns(group, sizePattern, nbPatterns, knowledgeGenerated)
    ax.clear()
    plot_instance(ax, problem)
    plot_patterns(ax, listPatterns, problem)
    return

#########################
# Core of the execution #
#########################

if __name__ == '__main__':
    # Read arguments
    allParameters = read_arguments(sys.argv[1:])
    
    # Configure the experiments for ONE instance
    instanceNumber = allParameters['instanceNumber']
    instanceType = allParameters['instanceType']
    instanceSize = allParameters['instanceSize'] 
    
    if allParameters['benchmark'] == "Solomon":
        subpath = os.path.join('Solomon', 'solomon_'+instanceSize, instanceType + instanceNumber)
    
    elif allParameters['benchmark'] == "Generated":
        if int(instanceNumber) < 10:
            instanceNumber = "0" + instanceNumber
        subpath = os.path.join('generated_instances', 'gen_'+instanceSize, instanceType + instanceNumber)

    job = configure_experiment(allParameters)
        
    # Run the study
    output_directory = 'data'
    experiment = Experiment(output_dir=output_directory, jobs=[job])
    experiment.run()

    output_path = os.path.join(output_directory, job.algorithm_tag, job.problem_tag)

    # Display of the knowledge 
    if allParameters["displayKnowledge"]:
        knowledgeGenerated = job.algorithm.knowledge
        maxPatternSize = job.algorithm.max_pattern_size
        numberOfGroups = job.algorithm.population_size - 1
        for i in range(0, numberOfGroups, 10):
            print("Aggregation of group ", i, ": ", job.algorithm.solutions[i].weights)
        fig, ax = plt.subplots()
        fig.subplots_adjust(bottom = 0.30)
    
        init_patternSize = 2
        init_group = 0
        init_nbPatterns = 10
        listPatterns = computePatterns(init_group, init_patternSize, init_nbPatterns, knowledgeGenerated)
        problem = job.algorithm.problem
        
        plot_instance(ax, problem)
        plot_patterns(ax, listPatterns, problem)

        axcolor = "White"
        
        # Set the axis
        group_axis = plt.axes([0.25, 0.2, 0.60, 0.03], 
                                facecolor = axcolor)
        patternSize_axis = plt.axes([0.25, 0.15, 0.60, 0.03], 
                                facecolor = axcolor)
        nbPattern_axis = plt.axes([0.25, 0.1, 0.60, 0.03], 
                                facecolor = axcolor)
        
        # Set the slider for the groups, the size of patterns and number of patterns
        group_slider = Slider(group_axis, 'Group', 
                                1, numberOfGroups, valinit = init_group, valstep = 1)
        size_slider = Slider(patternSize_axis, 'SizePatterns', 
                                2, maxPatternSize, valinit = init_patternSize, valstep = 1)
        nbPattern_slider = Slider(nbPattern_axis, 'Nb Patterns', 
                                1, 50, valinit = init_nbPatterns, valstep = 1)

        group_slider.on_changed(updateFigure)
        size_slider.on_changed(updateFigure)
        nbPattern_slider.on_changed(updateFigure)
        
        plt.show()
    