import logging
import io
import os
import threading
import time
import random as rd
from abc import abstractclassmethod, abstractmethod, ABC
from tracemalloc import take_snapshot
from typing import TypeVar, Generic, List
from jmetal.util.solution import get_non_dominated_solutions
from jmetal.core.quality_indicator import HyperVolume


from jmetal.config import store
from jmetal.core.problem import Problem
from jmetal.core.solution import FloatSolution, PermutationSolution
from jmetal.util.solution import print_function_values_to_file, print_variables_to_file, read_solutions

LOGGER = logging.getLogger('jmetal')
S = TypeVar('S')
R = TypeVar('R')

"""
.. module:: algorithm
   :platform: Unix, Windows
   :synopsis: Templates for algorithms.

.. moduleauthor:: Antonio J. Nebro <antonio@lcc.uma.es>, Antonio Benítez-Hidalgo <antonio.b@uma.es>
"""


class Algorithm(Generic[S, R], threading.Thread, ABC):

    def __init__(self, output_path, hypervolumeCriterion: float = 2.0):
        threading.Thread.__init__(self)
        self.external_archive = []
        self.solutions: List[S] = []
        self.evaluations = 0
        self.iterations = 0
        self.start_computing_time = 0
        self.total_computing_time = 0
        self.observable = store.default_observable
        self.output_path = output_path
        self.hvCriterion = hypervolumeCriterion


    @abstractmethod
    def create_initial_solutions(self) -> List[S]:
        """ Creates the initial list of solutions of a metaheuristic. """
        pass
    
    @abstractmethod
    def apply_local_search(self, solution: S, forced: bool) -> S:
        """ Applies a local search operator to improve the solution. """
        pass

    @abstractmethod
    def evaluate(self, solution_list: List[S]) -> List[S]:
        """ Evaluates a solution list. """
        pass

    @abstractmethod
    def init_progress(self) -> None:
        """ Initialize the algorithm. """
        pass

    @abstractmethod
    def stopping_condition_is_met(self) -> bool:
        """ The stopping condition is met or not. """
        pass

    @abstractmethod
    def step(self, external_archive: List[S] = []) -> None:
        """ Performs one iteration/step of the algorithm's loop. """
        pass

    @abstractmethod
    def update_progress(self) -> None:
        """ Update the progress after each iteration. """
        pass

    @abstractmethod
    def get_observable_data(self) -> dict:
        """ Get observable data, with the information that will be send to all observers each time. """
        pass

    @abstractmethod
    def get_hypervolume(self):
        pass

    def run(self):
        """ Execute the algorithm. """
        self.start_computing_time = time.time()

        self.solutions = self.create_initial_solutions()
        self.solutions = self.evaluate(self.solutions)
        
        n = len(self.solutions)

        LOGGER.debug('Initializing progress')
        self.init_progress()

        LOGGER.debug('Running main loop until termination criteria is met')
        ## Should move the current run to MOEAD directly and restore the original run !

        while (not self.stopping_condition_is_met()) and currentHV < self.hvCriterion:
            self.iterations += 1
            self.step()
            self.update_progress()
            if self.hvCriterion != 2.0:
                currentHV = self.get_hypervolume()
        self.total_computing_time = time.time() - self.start_computing_time


    @abstractmethod
    def get_result(self) -> R:
        pass

    @abstractmethod
    def get_name(self) -> str:
        pass


class DynamicAlgorithm(Algorithm[S, R], ABC):

    @abstractmethod
    def restart(self) -> None:
        pass


class EvolutionaryAlgorithm(Algorithm[S, R], ABC):

    def __init__(self,
                 problem: Problem[S],
                 population_size: int,
                 offspring_population_size: int,
                 output_path: str):
        super(EvolutionaryAlgorithm, self).__init__(output_path)
        self.problem = problem
        self.population_size = population_size
        self.offspring_population_size = offspring_population_size
        


    @abstractmethod
    def selection(self, population: List[S]) -> List[S]:
        """ Select the best-fit individuals for reproduction (parents). """
        pass

    @abstractmethod
    def reproduction(self, population: List[S]) -> List[S]:
        """ Breed new individuals through crossover and mutation operations to give birth to offspring. """
        pass

    @abstractmethod
    def replacement(self, population: List[S], offspring_population: List[S]) -> List[S]:
        """ Replace least-fit population with new individuals. """
        pass

    def get_observable_data(self) -> dict:
        return {'PROBLEM': self.problem,
                'EVALUATIONS': self.evaluations,
                'SOLUTIONS': self.get_result(),
                'COMPUTING_TIME': time.time() - self.start_computing_time}

    def init_progress(self) -> None:
        self.evaluations = self.population_size

        observable_data = self.get_observable_data()
        self.observable.notify_all(**observable_data)

    def step(self):

        N =self.problem.number_of_variables 

        mating_population = self.selection(self.solutions)
        offspring_population = self.reproduction(mating_population)
        offspring_population = self.evaluate(offspring_population)

        self.solutions = self.replacement(self.solutions, offspring_population)

    def analyse_injection(self):
        pass

    def update_progress(self) -> None:
        self.evaluations += self.offspring_population_size

        observable_data = self.get_observable_data()
        self.observable.notify_all(**observable_data)

    @property
    def label(self) -> str:
        return f'{self.get_name()}.{self.problem.get_name()}'

class LearningEvolutionaryAlgorithm(Algorithm[S, R], ABC):
    # only for online learning
    def __init__(self,
                 problem: Problem[S],
                 population_size: int,
                 offspring_population_size: int,
                 output_path: str, 
                 learningOnLO: bool,
                 hypervolumeCriterion: float = 2.0):
        super(LearningEvolutionaryAlgorithm, self).__init__(output_path, hypervolumeCriterion)

        self.problem = problem
        self.population_size = population_size
        self.offspring_population_size = offspring_population_size
        self.lsApplied = False
        self.learningOnLO = learningOnLO
        self.LS_time = 0
        self.Injection_time = 0
        self.Extraction_time = 0
        self.Archive_time = 0
        self.evaluation_time = 0

    @abstractmethod
    def selection(self, population: List[S]) -> List[S]:
        """ Select the best-fit individuals for reproduction (parents). """
        pass

    @abstractmethod
    def reproduction(self, population: List[S]) -> List[S]:
        """ Breed new individuals through crossover and mutation operations to give birth to offspring. """
        pass

    @abstractmethod
    def replacement(self, population: List[S], offspring_population: List[S]) -> List[S]:
        """ Replace least-fit population with new individuals. """
        pass

    @abstractmethod
    def extractKnowledge(self, population: List[S]) -> list:
        """ Extract the knowledge from the current population and update the existing one. """
        pass

    @abstractmethod
    def injectKnowledge(self, population: List[S]) -> List[S]:
        """ Inject the knowledge inside some individuals from the population. """
        pass

    def get_observable_data(self) -> dict:
        return {'PROBLEM': self.problem,
                'EVALUATIONS': self.evaluations,
                'SOLUTIONS': self.get_result(),
                'HYPERVOLUME': self.get_hypervolume(),
                'COMPUTING_TIME': time.time() - self.start_computing_time}

    def get_hypervolume(self) -> float:
        """ 
        Return the current hypervolume of the archive (self.external_archive). 
        Need to access to a file containing the extremas of the current instance. 
        Available for the MOVRPTW only.
        """
        indicator = HyperVolume([1.001, 1.001])
        normalizedObjectives = []
        
        # Use the extrema to normalize objectives
        for solution in self.external_archive:
            normalized = [round((solution.objectives[i] - self.problem.ideal[i])/max(1, (self.problem.nadir[i] - self.problem.ideal[i])),5) for i in range(len(solution.objectives))]
            normalizedObjectives.append(normalized)
        
        # Compute the hypervolume
        result = round(indicator.compute(normalizedObjectives), 3)
        return result
        
        
    def init_progress(self) -> None:
        self.evaluations = self.population_size

        observable_data = self.get_observable_data()
        self.observable.notify_all(**observable_data)


    def fast_insert(self, list: list, element):
        """ The list is sorted 
            Keep the list sorted at the end"""

        deb = 0
        fin = len(list)
        mid = (deb + fin)//2
        same = fin == 0
        found = False
        while not same:
            if list[mid].objectives > element.objectives:
                fin = mid
            elif list[mid].objectives < element.objectives:
                deb = mid
            else:
                found = True
                deb = mid
                fin = mid
            same = mid == (deb+fin)//2
            mid = (deb + fin)//2

        if not found:
            list.insert(fin, element)
        return list

    def __update_archive(self, solutions):
        for solution in solutions:
            self.fast_insert(self.external_archive, solution)

    def step(self):
        self.crossoverApplied = False
        self.injectionApplied = False
        self.lsApplied = False
        N =self.problem.number_of_variables 

        # Select solutions (neighbours)
        time_evaluate = time.time()
        mating_population = self.selection(self.solutions)
        
        # Crossover and then keep only one solution
        offspring_population = self.reproduction(mating_population)
        offspring_population = [offspring_population[rd.randint(0,1)]]
        
        # Evaluate the population
        offspring_population = self.evaluate(offspring_population)
        self.evaluation_time += time.time() - time_evaluate

        # Inject the knowledge
        time_injection = time.time()
        offspring_population = self.injectKnowledge(offspring_population)
        self.Injection_time += time.time() - time_injection

        # Apply Local Search (replace mutation)
        for s in offspring_population:
            time_ls = time.time()
            s = self.apply_local_search(s)
            self.LS_time += time.time() - time_ls

        # Update the neighborhood of the problem
        self.solutions = self.replacement(self.solutions, offspring_population)
        
        archive_time = time.time()
        self.__update_archive(self.solutions)
        self.external_archive = get_non_dominated_solutions(self.external_archive)
        self.Archive_time += time.time() - archive_time
        
        # Extraction of knowledge
        if self.lsApplied or not self.learningOnLO:
            time_extraction = time.time()
            self.knowledge = self.extractKnowledge(offspring_population)
            self.Extraction_time += time.time() - time_extraction

    def update_progress(self) -> None:
        self.evaluations += self.offspring_population_size

        observable_data = self.get_observable_data()
        self.observable.notify_all(**observable_data)

    @property
    def label(self) -> str:
        return f'{self.get_name()}.{self.problem.get_name()}'


class ParticleSwarmOptimization(Algorithm[FloatSolution, List[FloatSolution]], ABC):

    def __init__(self,
                 problem: Problem[S],
                 swarm_size: int):
        super(ParticleSwarmOptimization, self).__init__()
        self.problem = problem
        self.swarm_size = swarm_size

    @abstractmethod
    def initialize_velocity(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def initialize_particle_best(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def initialize_global_best(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def update_velocity(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def update_particle_best(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def update_global_best(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def update_position(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def perturbation(self, swarm: List[FloatSolution]) -> None:
        pass

    def get_observable_data(self) -> dict:
        return {'PROBLEM': self.problem,
                'EVALUATIONS': self.evaluations,
                'SOLUTIONS': self.get_result(),
                'COMPUTING_TIME': time.time() - self.start_computing_time}

    def init_progress(self) -> None:
        self.evaluations = self.swarm_size

        self.initialize_velocity(self.solutions)
        self.initialize_particle_best(self.solutions)
        self.initialize_global_best(self.solutions)

        observable_data = self.get_observable_data()
        self.observable.notify_all(**observable_data)

    def step(self):
        self.update_velocity(self.solutions)
        self.update_position(self.solutions)
        self.perturbation(self.solutions)
        self.solutions = self.evaluate(self.solutions)
        self.update_global_best(self.solutions)
        self.update_particle_best(self.solutions)

    def update_progress(self) -> None:
        self.evaluations += self.swarm_size

        observable_data = self.get_observable_data()
        self.observable.notify_all(**observable_data)

    @property
    def label(self) -> str:
        return f'{self.get_name()}.{self.problem.get_name()}'
