from abc import ABC, abstractmethod
from types import new_class
from typing import TypeVar, Generic, List

from jmetal.core.solution import PermutationSolution

S = TypeVar('S')
R = TypeVar('R')

"""
.. module:: Operator
   :platform: Unix, Windows
   :synopsis: Templates for operators.

.. moduleauthor:: Antonio J. Nebro <antonio@lcc.uma.es>, Antonio Benítez-Hidalgo <antonio.b@uma.es>
"""


class Operator(Generic[S, R], ABC):
    """ Class representing operator """

    @abstractmethod
    def execute(self, source: S) -> R:
        pass

    @abstractmethod
    def get_name(self) -> str:
        pass


def check_valid_probability_value(func):
    def func_wrapper(self, probability: float):
        if probability > 1.0:
            raise Exception('The probability is greater than one: {}'.format(probability))
        elif probability < 0.0:
            raise Exception('The probability is lower than zero: {}'.format(probability))

        res = func(self, probability)
        return res
    return func_wrapper


class Mutation(Operator[S, S], ABC):
    """ Class representing mutation operator. """

    @check_valid_probability_value
    def __init__(self, probability: float):
        self.probability = probability


class Crossover(Operator[List[S], List[R]], ABC):
    """ Class representing crossover operator. """

    @check_valid_probability_value
    def __init__(self, probability: float):
        self.probability = probability

    @abstractmethod
    def get_number_of_parents(self) -> int:
        pass

    @abstractmethod
    def get_number_of_children(self) -> int:
        pass


class Selection(Operator[S, R], ABC):
    """ Class representing selection operator. """

    def __init__(self):
        pass

class Extraction(Operator[S, R], ABC):
    """ Class representing extraction operator. """

    def __init__(self, probability):
        self.probability = probability

class Injection(Operator[S, S], ABC):
    """ Class representing injection operator. """

    def __init__(self, probability: float, strategy: str):
        self.probability = probability
        self.strategy = strategy

class LocalSearch(Operator[S, S], ABC):
    """ Class representing a local search operator for the MOVRPTW. """

    def __init__(self, probability = 1) -> None:
        self.probability = probability

    def generate_elements(self):
        """ Generate all possible candidates to start the operator. """
        pass

    def generate_neighborhood_element(self):
        """ Generate the neighborhood of an element. """
        pass
    
    def attributes_sequence(self, t_sequence, solution: PermutationSolution):
        # assert: sequence does not contain the depot
        sequence, revert = t_sequence
        indexes = (sequence[0], sequence[-1])
        if revert:
            attributes = solution.reverted_sequences[indexes[0]][indexes[1]]
        else:
            attributes = solution.sequences[indexes[0]][indexes[1]]
        return attributes

    def merge_2sequences(self, sequence1, attributes_sequence1, sequence2, attributes_sequence2, problem, solution, add_depot: bool):
        # assert: sequences never contain the depot and contain at least one element
        new_route = sequence1 + sequence2
        indexes_left = (sequence1[0], sequence1[-1])
        indexes_right = (sequence2[0], sequence2[-1])
        attributes_route = problem.concatenate_subsequences(indexes_left, attributes_sequence1, indexes_right, attributes_sequence2)
    
        if add_depot:
            attributes_route = problem.add_depot_to_sequence((new_route[0], new_route[-1]), attributes_route, solution)
        
        return new_route, attributes_route

    def merge_Nsequences(self, list_sequence, problem, solution):
        # Assert: sequence1 + sequence2 will never only return the depot
        # remove empty sequences:
        list_sequence = [i for i in list_sequence if i[0] != []]
        n = len(list_sequence)
        if n == 0:
            return [0,0], solution.sequences[0][0]
        
        if n == 1:
            (seq, revert) = list_sequence[0]
            merged_sequence = seq
            attributes_merged = self.attributes_sequence(list_sequence[0], solution)
            attributes_merged = problem.add_depot_to_sequence((merged_sequence[0], merged_sequence[-1]), attributes_merged, solution)
            merged_sequence = [0] + merged_sequence + [0]
            return merged_sequence, attributes_merged

        merged_sequence = list_sequence[0][0]
        attributes_merged = self.attributes_sequence(list_sequence[0], solution)

        for i in range(1, n):
            add_depot = i == n-1
            new_sequence = list_sequence[i]
            # now merge merged_sequence and new_sequence
            attributes_new_sequence = self.attributes_sequence(new_sequence, solution)
            merged_sequence, attributes_merged = self.merge_2sequences(merged_sequence, attributes_merged, new_sequence[0], attributes_new_sequence, problem, solution, add_depot)

        merged_sequence = [0] + merged_sequence + [0]
        return merged_sequence, attributes_merged

    def execute(self, element1, element2, problem, solution: S) -> R:
        routes = solution.structure
        (index_route1, index_u) = element1
        (index_route2, index_v) = element2
        route1 = routes[index_route1].copy()
        route2 = routes[index_route2].copy()

        indexes_R1 = (route1[1], route1[-2])
        attributes_R1 = solution.sequences[indexes_R1[0]][indexes_R1[1]]
        attributes_R1 = problem.add_depot_to_sequence(indexes_R1, attributes_R1, solution)

        indexes_R2 = (route2[1], route2[-2])
        attributes_R2 = solution.sequences[indexes_R2[0]][indexes_R2[1]]
        attributes_R2 = problem.add_depot_to_sequence(indexes_R2, attributes_R2, solution)

        new_route1, attributes_newR1, new_route2, attributes_newR2 = self.exchange_elements(element1, element2, problem, solution)

        if new_route2 == None:
            former_cost = attributes_R1['C']
            former_wt = attributes_R1['WT']
            new_cost = attributes_newR1['C']
            new_wt = attributes_newR1['WT']
            check_delay = attributes_newR1['TW'] == 0
            check_capacity = attributes_newR1['Q'] <= problem.capacity
        else:
            former_cost = attributes_R1['C'] + attributes_R2['C']
            former_wt = attributes_R1['WT'] + attributes_R2['WT']
            new_cost = attributes_newR1['C'] + attributes_newR2['C']
            new_wt = attributes_newR1['WT'] + attributes_newR2['WT']
            check_delay = attributes_newR1['TW'] + attributes_newR2['TW'] == 0
            check_capacity = attributes_newR1['Q'] + attributes_newR2['Q'] <= problem.capacity 

        current_fitness = solution.weights[0] * former_cost + solution.weights[1] * former_wt
        new_fitness = solution.weights[0] * new_cost + solution.weights[1] * new_wt

        if check_capacity and check_delay:
            if round(new_fitness, 5) < round(current_fitness, 5):
                if new_route2 == None:
                    routes_modified = [index_route1]
                    new_routes = [new_route1.copy()]
                else:
                    routes_modified = [index_route1, index_route2]
                    new_routes = [new_route1.copy(), new_route2.copy()]
                
                objective0 = solution.objectives[0] - former_cost + new_cost
                objective1 = solution.objectives[1] - former_wt + new_wt

                objectives = [objective0, objective1]
                return routes_modified, new_routes, objectives 
        return None, None, None 

    def apply_move(self, index_route_modified: list, new_routes: list, objectives, problem, solution):
        """ Since the move has already been checked, we directly apply it. """
        routes = solution.structure
        n = len(index_route_modified)
        nb_to_remove = 0
        for i in range(n):
            index_route = index_route_modified[i]
            new_route = new_routes[i]
            routes[index_route] = new_route.copy()
            if new_route == [0,0]:
                nb_to_remove += 1
            else:
                problem.compute_subsequences(new_route[1:-1], solution, reverse = False)
                reverted_new_route = new_route.copy()
                reverted_new_route.reverse()
                problem.compute_subsequences(reverted_new_route[1:-1], solution, reverse= True)

        # Remove empty routes 
        for _ in range(nb_to_remove):
            routes.remove([0,0])

        # Update the list of variables
        new_variables = []
        for r in routes:
            for c in r:
                if c != 0:
                    new_variables.append(c-1)
        
        solution.variables = new_variables.copy()
        solution.objectives[0] = objectives[0]
        solution.objectives[1] = objectives[1]
        solution.improvedByLS = True
        return solution