import copy
import random
from re import sub
from typing import List

from jmetal.core.operator import Injection
from jmetal.core.solution import Solution,  PermutationSolution
from jmetal.problem.multiobjective.movrptw import MOVRPTW
from jmetal.util.ckecking import Check
from jmetal.core.problem import Problem

"""
.. module:: extraction
   :platform: Unix, Windows
   :synopsis: Module implementing crossover operators.

.. moduleauthor:: Clément Legrand <clement.legrand4.etu@univ-lille.fr>
"""

class MOPatternInjection(Injection[PermutationSolution]):
    def __init__(self, probability, strategy):
        super(MOPatternInjection, self).__init__(probability, strategy)
        self.R_best = [] # global variable for best-reconnect
        self.bestCost = 10**10
        self.improved = 0

    def cost_of_fragment(self, problem, solution, attributes_fragment):
        return 5000*max(attributes_fragment['Q']-problem.capacity,0) + \
                solution.weights[0]*attributes_fragment['C'] + \
                solution.weights[1]*attributes_fragment['WT'] + \
                50000*attributes_fragment['TW']

    def information_fragment(self, problem: MOVRPTW, fragment, solution: PermutationSolution):
        """
        Compute the attributes of a fragment of solution (duration, cost, waiting time...)
        WARNING: the solution MUST have solution.sequences up to date
        """
        c = 0
        d = 0
        delayTime = 0
        waitingTime = 0

        if len(fragment) == 1:
            attributes_route = solution.sequences[fragment[0]][fragment[0]]
        elif fragment[0] == 0 and fragment[-1] == 0:
            attributes_route = solution.sequences[fragment[1]][fragment[-2]]
            attributes_route = problem.add_depot_to_sequence((fragment[1], fragment[-2]), attributes_route, solution)
        elif fragment[0] == 0:
            attributes_route = solution.sequences[fragment[1]][fragment[-1]]
            attributes_route = problem.add_depot_to_left((fragment[1], fragment[-1]), attributes_route, solution)
        elif fragment[-1] == 0:
            attributes_route = solution.sequences[fragment[0]][fragment[-2]]
            attributes_route = problem.add_depot_to_right((fragment[0], fragment[-2]), attributes_route, solution)
        else:
            attributes_route = solution.sequences[fragment[0]][fragment[-1]]
            
        waitingTime += attributes_route['WT']
        c += attributes_route['C']
        d += attributes_route['Q']
        delayTime += attributes_route['TW']
        return attributes_route    

    def cost_of_list_of_fragments(self, problem, solution, list_fragments):
        total_cost = 0
        for (_, attributes) in list_fragments:
            total_cost += self.cost_of_fragment(problem, solution, attributes)
        return total_cost

    def check_pattern_present(self, solution, pattern):
        l = len(pattern)
        list_routes = solution.structure
        start_pattern = pattern[0]
        is_present = False
        for r in list_routes:
            if start_pattern in list_routes:
                i = r.index(start_pattern)
                if i+l-1 < len(r):
                    is_present = pattern == r[i:i+l]
            break
        return is_present

    def cut_into_fragments(self, problem, solution, pattern):
        list_routes = solution.structure
        R_inv = [] # routes which not contain any customers in the pattern and thus are not modified
        R_init = [] # routes which contain at least one customer of pattern
        R_beg = [] # fragments of routes which start with the depot
        R_mid = [] # fragments of routes which don't contain the depot
        R_end = [] # fragments of routes which end with the depot

        for r in list_routes:
            frag = []
            inv = True
            for c in r:
                if c in pattern:
                    inv = False
                    if frag != []:
                        # Try to optimize -> define a global variable to store computed frags with their costs
                        attributes_fragment = self.information_fragment(problem, frag, solution)
                        if frag[0] == 0:
                            R_beg.append((frag, attributes_fragment))
                            
                        else:
                            R_mid.append((frag, attributes_fragment))
                        frag = []
                else:
                    frag.append(c)
            if inv:
                attributes_fragment = self.information_fragment(problem, r, solution)
                R_inv.append((r,attributes_fragment))
            else:
                attributes_fragment = self.information_fragment(problem, r, solution)
                R_init.append((r,attributes_fragment))
                if frag != []:
                    attributes_fragment = self.information_fragment(problem, frag, solution)
                    R_end.append((frag, attributes_fragment))

        n = len(pattern)
        attributes_pattern = solution.sequences[pattern[0]][pattern[0]]
        current_pattern = [pattern[0]]
        for i in range(1, n):
            next_customer = pattern[i]
            attributes_pattern = problem.concatenate_subsequences((current_pattern[0], current_pattern[-1]), attributes_pattern, (next_customer, next_customer), solution.sequences[next_customer][next_customer])
            current_pattern.append(next_customer)
        R_mid.append((pattern, attributes_pattern))
        self.R_best = R_init.copy()

        return R_inv, R_beg, R_mid, R_end

    def best_reconnect(self, problem: MOVRPTW, solution: PermutationSolution, beg, mid, end, complete):
        """
        best-reconnect the pieces of routes (similar to PILS algo)
        """
        (R_beg, cost_beg) = beg
        (R_mid, cost_mid) = mid
        (R_end, cost_end) = end
        (R_complete, cost_complete) = complete

        totalCost = cost_beg + cost_mid + cost_end + cost_complete

        if totalCost < self.bestCost:
            if len(R_beg) == 0:
                self.R_best = R_complete.copy()
                solution.improvedByInjection += 1
                self.improved = 1
                self.bestCost = totalCost
            else:
                f_beg = random.sample(R_beg, 1)[0]
                cost_deleted = self.cost_of_list_of_fragments(problem, solution, [f_beg])

                for f_mid in R_mid:
                    new_R_beg = R_beg.copy()
                    new_R_beg.remove(f_beg)
                    new_R_mid = R_mid.copy()
                    new_R_mid.remove(f_mid)

                    new_costMid = cost_mid - self.cost_of_list_of_fragments(problem, solution, [f_mid])
                    new_mid = (new_R_mid.copy(), new_costMid)

                    new_frag = f_beg[0] + f_mid[0] # the new fragment (starts with depot)
                    attributes_frag = problem.concatenate_subsequences((f_beg[0][0], f_beg[0][-1]), f_beg[1], (f_mid[0][0], f_mid[0][-1]), f_mid[1])

                    new_R_beg.append((new_frag, attributes_frag))

                    cost_added = self.cost_of_list_of_fragments(problem, solution, [new_R_beg[-1]])
                    new_costBeg = cost_beg - cost_deleted + cost_added
                    new_beg = (new_R_beg.copy(), new_costBeg)
                    new_end = (R_end.copy(), cost_end)
                    new_complete = (R_complete.copy(), cost_complete)

                    self.best_reconnect(problem, solution, new_beg, new_mid, new_end, new_complete)

                if len(R_beg) != 1 or len(R_mid) == 0:
                    for f_end in R_end:
                        new_R_beg = R_beg.copy()
                        new_R_beg.remove(f_beg)
                        new_R_end = R_end.copy()
                        new_R_end.remove(f_end)

                        new_costEnd = cost_end - self.cost_of_list_of_fragments(problem, solution, [f_end])
                        new_end = (new_R_end.copy(), new_costEnd)
                        new_costBeg = cost_beg - cost_deleted
                        new_beg = (new_R_beg.copy(), new_costBeg)


                        new_frag = f_beg[0] + f_end[0] # the new fragment (starts with depot)
                        attributes_frag = problem.concatenate_subsequences((f_beg[0][0], f_beg[0][-1]), f_beg[1], (f_end[0][0], f_end[0][-1]), f_end[1])

                        new_R = R_complete.copy()
                        new_R.append((new_frag, attributes_frag))

                        new_costR = cost_complete + self.cost_of_list_of_fragments(problem, solution, [new_R[-1]])
                        new_complete = (new_R.copy(), new_costR)

                        new_mid = (R_mid.copy(), cost_mid)

                        self.best_reconnect(problem, solution, new_beg, new_mid, new_end, new_complete)

    def inject_one_pattern_pils(self, problem: MOVRPTW, solution: PermutationSolution, pattern):
        """
        Try to inject the pattern into the solution by using the injection of PILS method.
        Weights are used to compute the fitness of the solution
        """
        # verify that the pattern is not already present in the solution
        if not self.check_pattern_present(solution, pattern):
        
            # first cut the routes of the solution into fragments
            R_inv, R_beg, R_mid, R_end = self.cut_into_fragments(problem, solution, pattern)
            cost_INV = self.cost_of_list_of_fragments(problem, solution, R_inv)
            self.bestCost = solution.weights[0] * solution.objectives[0] + solution.weights[1] * solution.objectives[1] - cost_INV
            cost_beg = self.cost_of_list_of_fragments(problem, solution, R_beg)
            cost_mid = self.cost_of_list_of_fragments(problem, solution, R_mid)
            cost_end = self.cost_of_list_of_fragments(problem, solution, R_end)
            
            # apply best-reconnect to merge R_beg, R_mid and R_end
            self.best_reconnect(problem, solution, (R_beg, cost_beg), (R_mid, cost_mid), (R_end, cost_end), ([], 0))

            # R_best contains the reconnected routes
            solution.objectives[0] = 0
            solution.objectives[1] = 0
            new_routes = []
            total_delay = 0
            all_routes = R_inv + self.R_best
            s = []

            # Update the sequences
            for (r, _) in self.R_best: # routes modified
                problem.compute_subsequences(r[1:-1], solution, False)
                reversed_r = r.copy()
                reversed_r.reverse()
                problem.compute_subsequences(reversed_r[1:-1], solution, True)
            
            for (i, att) in all_routes:
                if i != [0,0]:
                    for c in i:
                        if c != 0:
                            s.append(c-1)


                    attributes_i = solution.sequences[i[1]][i[-2]]
                    attributes_i = problem.add_depot_to_sequence((i[1], i[-2]), attributes_i, solution)
                    solution.objectives[0] += attributes_i['C'] 
                    solution.objectives[1] += attributes_i['WT'] 
                    total_delay += attributes_i['TW'] 
                    new_routes.append(i.copy())

            solution.variables = s.copy()
            solution.structure = copy.deepcopy(new_routes)
        return solution

    def execute(self, problem: Problem, solution:PermutationSolution, chosen_patterns: list) -> PermutationSolution:
        """
        Inject patterns from chosen_patterns one by one. The pattern is kept only when the solution is improved
        """
        
        if random.random() > self.probability:
            return solution

        tour = [0] + [v + 1 for v in solution.variables]
        problem.compute_subsequences(tour, solution, reverse = False)
        for pattern in chosen_patterns:
            solution = self.inject_one_pattern_pils(problem, solution, pattern)
        return solution

    def get_name(self):
        return 'Multi-Objective pattern injection'
