import copy
import random
import time
from typing import List

from numpy.random.mtrand import noncentral_chisquare, rand

from jmetal.core.operator import LocalSearch
from jmetal.core.solution import Solution,  PermutationSolution
from jmetal.operator import crossover
from jmetal.problem.multiobjective.movrptw import MOVRPTW
from jmetal.util.ckecking import Check
from jmetal.core.problem import Problem
from jmetal.util.neighborhood import Neighborhood

"""
.. module:: extraction
   :platform: Unix, Windows
   :synopsis: Module implementing local search operators.

.. moduleauthor:: Clément Legrand <clement.legrand4.etu@univ-lille.fr>
"""

class TwoOptStar(LocalSearch[PermutationSolution]):
    """ Inter-route move. Exchange the tails of two distinct routes. 
        Random exploration, first improvement. """

    def generate_elements(self, solution: PermutationSolution):
        """ Return a permutation of all possible elements to realise the move. """
        routes = solution.structure
        elements = []
        for i in range(len(routes)):
            current_route = routes[i]
            elements += [(i,j) for j in range(len(current_route)-1)] # possible to remove the depot
        random.shuffle(elements)
        return elements

    def generate_neighborhood_element(self, element: tuple, possible_neighbours: list, solution: PermutationSolution):
        """ An arc is completely characterised by the index of the route and the position where starts the arc. 
            e.g. (2,3) refers to the arc from route 2 which starts at position 3 (in the route).  
        """
        routes = solution.structure
        neighborhood_arc = []
        for i in range(len(routes)):
            if i != element[0]:
                current_route = routes[i]
                neighborhood_arc += [(i,j) for j in range(len(current_route)-1) if possible_neighbours[current_route[j]]]
        random.shuffle(neighborhood_arc)
        return neighborhood_arc

    def exchange_elements(self, element1: tuple, element2: tuple, problem: MOVRPTW, solution: PermutationSolution):
        (indexR1, arc1) = element1
        (indexR2, arc2) = element2
        routes = solution.structure
        route1, route2 = routes[indexR1], routes[indexR2]
        
        list_sequences1 = [(route1[1:arc1+1], False), (route2[arc2+1:-1], False)]
        list_sequences2 = [(route2[1:arc2+1], False), (route1[arc1+1:-1], False)]
        
        new_route1, attributes_newR1 = self.merge_Nsequences(list_sequences1, problem, solution)
        new_route2, attributes_newR2 = self.merge_Nsequences(list_sequences2, problem, solution)
        return new_route1, attributes_newR1, new_route2, attributes_newR2
    
    def get_name(self) -> str:
        return "2-opt*"

class Swap(LocalSearch[PermutationSolution]):
    def generate_elements(self, solution: PermutationSolution):
        routes = solution.structure
        elements = []
        for r in range(len(routes)):
            current_route = routes[r]
            elements += [(r, i) for i in range(1, len(current_route)-1)]
        random.shuffle(elements)
        return elements

    def generate_neighborhood_element(self, element: tuple, possible_neighbours: list, solution: PermutationSolution):
        routes = solution.structure
        neighbours = []
        for r in range(len(routes)):
            current_route = routes[r]
            neighbours += [(r, i) for i in range(1, len(current_route)-1) if possible_neighbours[current_route[i]] and (r,i) != element]
        random.shuffle(neighbours)
        return neighbours

    def exchange_elements(self, element1: tuple, element2: tuple, problem: MOVRPTW, solution: PermutationSolution):
        routes = solution.structure
        (index_r1, index_u) = element1
        (index_r2, index_v) = element2

        route1 = routes[index_r1]
        route2 = routes[index_r2]

        if index_r1 == index_r2:
            min_index = min(index_v, index_u)
            max_index = max(index_v, index_u)
            if route1[min_index+1:max_index] == []:
                list_sequences1 = [(route1[1:min_index], False), ([route1[max_index]], False), ([route1[min_index]], False), (route1[max_index+1:-1], False)]
            else:
                list_sequences1 = [(route1[1:min_index], False), ([route1[max_index]], False), (route1[min_index+1:max_index], False), ([route1[min_index]], False), (route1[max_index+1:-1], False)]

            new_route1, attributes_newR1 = self.merge_Nsequences(list_sequences1, problem, solution)
            new_route2 = None
            attributes_newR2 = None
        else:
            list_sequences1 = [(route1[1:index_u], False), ([route2[index_v]], False), (route1[index_u+1:-1], False)]
            list_sequences2 = [(route2[1:index_v], False), ([route1[index_u]], False), (route2[index_v+1:-1], False)]
            new_route1, attributes_newR1 = self.merge_Nsequences(list_sequences1, problem, solution)
            new_route2, attributes_newR2 = self.merge_Nsequences(list_sequences2, problem, solution)

        return new_route1, attributes_newR1, new_route2, attributes_newR2 

    def get_name(self) -> str:
        return "Swap"

class Relocate(LocalSearch[PermutationSolution]):
    def generate_elements(self, solution: PermutationSolution):
        routes = solution.structure
        elements = []
        for r in range(len(routes)):
            current_route = routes[r]
            elements += [(r, i) for i in range(1, len(current_route)-1)]
        random.shuffle(elements)
        return elements

    def generate_neighborhood_element(self, element: tuple, possible_neighbours: list, solution: PermutationSolution):
        routes = solution.structure
        neighbours = []
        for r in range(len(routes)):
            current_route = routes[r]
            if r != element[0]:
                neighbours += [(r, i) for i in range(1, len(current_route)-1) if possible_neighbours[current_route[i]]] #or possible_neighbours[current_route[i+1]]]
            else:
                neighbours += [(r, i) for i in range(1, len(current_route)-1) if possible_neighbours[current_route[i]] and (r, i+1) != element and (r, i) != element and len(current_route) > 3]
        random.shuffle(neighbours)
        return neighbours

    def exchange_elements(self, element1: tuple, element2: tuple, problem: MOVRPTW, solution: PermutationSolution):
        """ Put u after v. """
        routes = solution.structure
        (index_r1, index_u) = element1
        (index_r2, index_v) = element2
        
        route1 = routes[index_r1]
        route2 = routes[index_r2]

        if index_r1 == index_r2:
            if index_v < index_u:
                list_sequences1 = [(route1[1:index_v+1], False), ([route1[index_u]], False), (route1[index_v+1:index_u], False), (route1[index_u+1:-1], False)]
            else:
                list_sequences1 = [(route1[1:index_u], False), (route1[index_u+1:index_v+1], False), ([route1[index_u]], False), (route1[index_v+1:-1], False)]

            new_route1, attributes_newR1 = self.merge_Nsequences(list_sequences1, problem, solution)
            new_route2 = None
            attributes_newR2 = None

        else:
            list_sequences1 = [(route1[1:index_u], False), (route1[index_u+1:-1], False)]
            list_sequences2 = [(route2[1:index_v+1], False), ([route1[index_u]], False), (route2[index_v+1:-1], False)]
            new_route1, attributes_newR1 = self.merge_Nsequences(list_sequences1, problem, solution)
            new_route2, attributes_newR2 = self.merge_Nsequences(list_sequences2, problem, solution)

        return new_route1, attributes_newR1, new_route2, attributes_newR2 

    def get_name(self) -> str:
        return "Relocate"

class ManyOperators(TwoOptStar, Swap, Relocate):
    """
    Apply many LS operators to reach a local optimum
    """

    def __init__(self, probability) -> None:
        super(ManyOperators, self).__init__(probability)

    def execute(self, problem: MOVRPTW, solution: PermutationSolution) -> PermutationSolution:
        x = random.random()
        
        if x > self.probability:
            return solution, 0

        twoOptStar = TwoOptStar()
        swap = Swap()
        relocate = Relocate()

        mode_ls = "BEST" # Choose between "FIRST" and "BEST"

        operators = [twoOptStar, swap, relocate]

        random.shuffle(operators)
        for operator in operators:
            solution.improvedByLS = True
            nb_iteration = 0

            time_ope = time.time()
            size_neighborhood = 0

            while solution.improvedByLS and nb_iteration < 1000:
                nb_iteration += 1

                best_fitness = 10**16
                best_move = None
                solution.improvedByLS = False
                elements_to_consider = operator.generate_elements(solution)
                k = len(elements_to_consider)
                cpt_out = 0

                # find the first element for which we obtain an improvement
                while not solution.improvedByLS and cpt_out < k:
                    current_element = elements_to_consider[cpt_out] 
                    current_customer = current_element[1]
                    possible_neighbours = problem.binary_neighbours[current_customer]
                    neighborhood_element = operator.generate_neighborhood_element(current_element, possible_neighbours, solution)
                    size_neighborhood += len(neighborhood_element)
                    l = len(neighborhood_element)

                    # Find the best neighbor for the current_element
                    if mode_ls == "FIRST":
                        cpt_in = 0
                        while not solution.improvedByLS and cpt_in < l:
                            neighbor = neighborhood_element[cpt_in]
                            cpt_in += 1
                            result = operator.execute(current_element, neighbor, problem, solution)
                            if result != (None, None, None):
                                solution = operator.apply_move(result[0], result[1], result[2], problem, solution)
                        
                    elif mode_ls == "BEST":
                        for neighbor in neighborhood_element:
                            result = operator.execute(current_element, neighbor, problem, solution)
                            if result != (None, None, None):
                                index_routes_modified, new_routes, objectives = result
                                new_fitness = solution.weights[0] * objectives[0] + solution.weights[1] * objectives[1]
                                if round(new_fitness, 2) < round(best_fitness, 2):
                                    best_fitness = new_fitness
                                    best_move = result

                        # Apply the best move found
                        if best_move != None:
                            solution = operator.apply_move(best_move[0], best_move[1], best_move[2], problem, solution)
                    
                    cpt_out += 1
            time_ope = time.time() - time_ope
        return solution, 1

    def get_name(self) -> str:
        return "Local Search: First-Best"